# -*- coding: utf-8 -*-
import redis
import pickle
from .settings import REDIS_HOST, REDIS_PORT, REDIS_CHANNELS

__all__ = ('publish',)

r = redis.Redis(host=REDIS_HOST, port=REDIS_PORT)


def publish(channels_list=None, *args, **kwargs):
    """
    Publishes a message on channels
    """
    channels_list = channels_list or REDIS_CHANNELS
    p = pickle.dumps((args, kwargs), protocol=2)

    if not isinstance(channels_list, list):
        channels_list = [channels_list]

    for channel in channels_list:
        r.publish(channel, p)



# -*- coding: utf-8 -*-


class BaseSubscriber(object):
    """
    Base subscriber that dispatch message to callback functions
    """
    def dispatch_message(self, channel, args, kwargs):
        """
        Calls callback functions
        """
        for func, channels in self.application.registry.senders:
            if channels is None or channel in channels:
                self.application.executor.submit(
                    func, channel, self.application.clients, *args, **kwargs)



# -*- coding: utf-8 -*-
import json
import logging
import tornado.gen
import tornado.ioloop
import tornado.web
import tornado.websocket
from toredis import Client
from concurrent.futures import ThreadPoolExecutor
from django.utils.module_loading import import_by_path

from .settings import SUBSCRIBER_CLASS_PATH, REDIS_HOST, REDIS_PORT
from .settings import REDIS_CHANNELS, MAX_WORKERS
from .registry import registry, autodiscover

autodiscover()

__all__ = ('WebSocketServer',)
LOGGER = logging.getLogger('mease.websocket_server')
SUBSCRIBER_CLASS = import_by_path(SUBSCRIBER_CLASS_PATH)


class WebSocketHandler(tornado.websocket.WebSocketHandler):

    def open(self):
        """
        Called when a client opens a websocket connection
        """
        # Init storage
        self.storage = {}

        # Call openers callbacks
        for func in self.application.registry.openers:
            self.application.executor.submit(func, self, self.application.clients)

        # Append client to clients list
        if self not in self.application.clients:
            self.application.clients.append(self)

    def on_close(self):
        """
        Called when a client closes a websocket connection
        """
        # Call closer callbacks
        for func in self.application.registry.closers:
            self.application.executor.submit(func, self, self.application.clients)

        # Remove client from clients list
        if self in self.application.clients:
            self.application.clients.remove(self)

    def on_message(self, message):
        """
        Called when a client sends a message through websocket
        """
        # Parse JSON
        try:
            json_message = json.loads(message)
        except ValueError:
            json_message = None

        # Call receiver callbacks
        for func, to_json in self.application.registry.receivers:

            if to_json:
                if json_message is None:
                    continue
                msg = json_message
            else:
                msg = message

            self.application.executor.submit(
                func, self, msg, self.application.clients)

    def send(self, *args, **kwargs):
        self.write_message(*args, **kwargs)


class WebSocketServer(SUBSCRIBER_CLASS):

    def __init__(self, debug, port):

        # Tornado loop
        self.ioloop = tornado.ioloop.IOLoop.instance()

        # Tornado application
        self.debug = debug
        self.port = port

        self.application = tornado.web.Application([
            (r'/', WebSocketHandler),
        ], debug=self.debug)

        # Init application storage
        self.application.storage = {}

        # Clients list
        self.application.clients = []

        # Registry
        self.application.registry = registry

        # Executor
        self.application.executor = ThreadPoolExecutor(max_workers=MAX_WORKERS)

        # Connect to redis
        LOGGER.info("Connecting to Redis on {host}:{port}".format(
            host=REDIS_HOST, port=REDIS_PORT))

        self.redis_client = Client()
        self.redis_client.connect(host=REDIS_HOST, port=REDIS_PORT)

        LOGGER.info("Successfully connected to Redis")

        # Subscribe to channels
        self.redis_client.subscribe(REDIS_CHANNELS, callback=self.on_message)

        LOGGER.debug("Subscribed to [{channels}] Redis channels".format(
            channels=', '.join(REDIS_CHANNELS)))

    def run(self):
        """
        Starts websocket server
        """
        self.application.listen(self.port)
        self.ioloop.start()



# -*- coding: utf-8 -*-
from .registry import registry

__all__ = ('opener', 'closer', 'receiver', 'sender')


def opener(function=None):
    def actual_decorator(f):
        registry.opener(f)

    if function:
        return actual_decorator(function)
    return actual_decorator


def closer(function=None):
    def actual_decorator(f):
        registry.closer(f)

    if function:
        return actual_decorator(function)
    return actual_decorator


def receiver(function=None, json=False):
    def actual_decorator(f):
        registry.receiver(f, json)

    if function:
        return actual_decorator(function)
    return actual_decorator


def sender(function=None, channels=None):
    def actual_decorator(f):
        registry.sender(f, channels)

    if function:
        return actual_decorator(function)
    return actual_decorator



# -*- coding: utf-8 -*-
import pickle

from .base import BaseSubscriber


class RedisSubscriber(BaseSubscriber):

    def on_message(self, message):
        """
        Redis pubsub callback
        """
        event, channel, data = message

        if event.decode() == 'message':
            args, kwargs = pickle.loads(data)
            self.dispatch_message(channel.decode(), args, kwargs)



# -*- coding: utf-8 -*-
import logging
from django.core.management.base import BaseCommand
from django.conf import settings
from optparse import make_option

from ...server import WebSocketServer

DEBUG = getattr(settings, 'DEBUG', False)
LOGGER = logging.getLogger('mease.websocket_server')


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        # Debug
        make_option(
            '-d',
            '--debug',
            action='store_true',
            dest='debug',
            default=DEBUG,
            help="Debug mode"),

        # Listen port
        make_option(
            '-p',
            '--port',
            action='store',
            dest='port',
            default=9090,
            help="Port to listen"),
    )

    def _registry_names(self, registry):
        """
        Returns functions names for a registry
        """
        return ', '.join(
            f.__name__ if not isinstance(f, tuple) else f[0].__name__
            for f in registry)

    def handle(self, *args, **options):
        """
        Starts websocket server
        """
        debug, port = options['debug'], options['port']

        websocket_server = WebSocketServer(
            debug=options['debug'], port=options['port'])

        # Log registry
        LOGGER.debug("Registered callback functions :")

        LOGGER.debug(
            "Openers : [%s]" % self._registry_names(
                websocket_server.application.registry.openers))

        LOGGER.debug(
            "Closers : [%s]" % self._registry_names(
                websocket_server.application.registry.closers))

        LOGGER.debug(
            "Receivers : [%s]" % self._registry_names(
                websocket_server.application.registry.receivers))

        LOGGER.debug(
            "Senders : [%s]" % self._registry_names(
                websocket_server.application.registry.senders))

        # Start server
        LOGGER.info("Started websocket server on port {port}".format(port=port))
        websocket_server.run()



# -*- coding: utf-8 -*-
from django.conf import settings
from django.utils.importlib import import_module

__all__ = ('autodiscover', 'registry', 'opener', 'closer', 'receiver', 'sender')


def autodiscover():
    """
    Autodiscovers apps with `mease_registry` modules
    """
    for app in getattr(settings, 'INSTALLED_APPS', []):
        import_module(app)

        try:
            import_module('%s.mease_registry' % app)
        except ImportError:
            pass


class MeaseRegistry(object):
    """
    Registry for mease callbacks
    """
    def __init__(self):
        self.openers = []
        self.closers = []
        self.receivers = []
        self.senders = []

    def opener(self, func):
        """
        Registers an opener function
        """
        self.openers.append(func)

    def closer(self, func):
        """
        Registers a closer function
        """
        self.closers.append(func)

    def receiver(self, func, json=False):
        """
        Registers a receiver function
        """
        self.receivers.append((func, json))

    def sender(self, func, channels=None):
        """
        Registers a sender function
        """
        self.senders.append((func, channels))

registry = MeaseRegistry()



import threading
from django.core.management import call_command
from django.test import TestCase

from .settings import TEST_WEBSOCKET_PORT


threading.Thread(
    target=call_command,
    args=['start_websocket_server'],
    kwargs={'port': TEST_WEBSOCKET_PORT, 'host': 'localhost'})


class MeaseTestCase(TestCase):
    def test_dicks(self):
        import time
        time.sleep(10)



# -*- coding: utf-8 -*-
from django.conf import settings

__all__ = (
    'SUBSCRIBER_CLASS_PATH', 'REDIS_CHANNELS', 'REDIS_HOST', 'REDIS_PORT', 'MAX_WORKERS')

SUBSCRIBER_CLASS_PATH = getattr(
    settings, 'MEASE_SUBSCRIBER_CLASS', 'mease.subscribers.redis.RedisSubscriber')

REDIS_CHANNELS = getattr(settings, 'MEASE_REDIS_CHANNELS', 'websocket')
REDIS_HOST = getattr(settings, 'MEASE_REDIS_HOST', 'localhost')
REDIS_PORT = getattr(settings, 'MEASE_REDIS_PORT', 6379)

MAX_WORKERS = getattr(settings, 'MEASE_MAX_WORKERS', 15)
